/*
 * Copyright (c) 2010 Martin Aigner, Andreas Haas, Stephanie Stroka
 * http://cs.uni-salzburg.at/~maigner
 * http://cs.uni-salzburg.at/~ahaas
 * http://cs.uni-salzburg.at/~sstroka
 *
 * University Salzburg, www.uni-salzburg.at
 * Department of Computer Science, cs.uni-salzburg.at
 */

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <pthread.h>
#include <string.h>
#include "stm.h"
#include "scm-desc.h"
#include "regmalloc.h"
#include "meter.h"
#include "arch.h"

#ifdef SCM_PRINTMEM
#include <malloc.h>
#endif //SCM_PRINTMEM

#ifdef SCM_METER
#include "tracer.h"
#endif

static long global_time = 0;
static unsigned int number_of_threads = 0;

//the number of threads, that have not yet ticked in a global period
static unsigned int ticked_threads_countdown = 1;

//protects global_time, number_of_threads and ticked_threads_countdown
static pthread_spinlock_t global_time_lock;

static descriptor_root_t *terminated_descriptor_roots = NULL;

//protects the data structures of terminated threads
static pthread_mutex_t terminated_descriptor_roots_mutex =
    PTHREAD_MUTEX_INITIALIZER;

//the first thread joining libscm has to initialize the thread local storage
static unsigned int init_threads = 0;
static pthread_key_t descriptor_root_key;

static inline void lock_global_time();
static inline void unlock_global_time();
static inline void lock_descriptor_roots();
static inline void unlock_descriptor_roots();
static descriptor_root_t *new_descriptor_root();
static descriptor_root_t *scm_register_thread();
static void increment_and_expire_clock(const unsigned long clock);

/**
 * Creates and initializes a new region page if no other
 * region page exists or if all other region pages are full.
 * The region_page is allocated page-aligned.
 */
static region_page_t* init_region_page(region_t* region);

void *__wrap_malloc(size_t size);
void *__wrap_calloc(size_t nelem, size_t elsize);
void *__wrap_realloc(void *ptr, size_t size);
void __wrap_free(void *ptr);
size_t __wrap_malloc_usable_size(void *ptr);

//avoid ELF interposition of exported but internally used symbols
//by creating static aliases
static void scm_resume_thread_internal(void)
__attribute__((alias("scm_resume_thread")));

static void scm_block_thread_internal(void)
__attribute__((alias("scm_block_thread")));

static void* __wrap_malloc_internal(size_t size)
__attribute__((alias("__wrap_malloc")));

void* scm_malloc(size_t size)
__attribute__((alias("__wrap_malloc")));

void scm_free(void *ptr)
__attribute__((alias("__wrap_free")));

#ifndef SCM_EAGER_COLLECTION
/**
 * Collects descriptors incrementally
 */
static void scm_lazy_collect(void);
#else
/**
 * Collects descriptors all at once
 */
static void scm_eager_collect(void);
#endif

void *__wrap_calloc(size_t nelem, size_t elsize) {

    void *p = __wrap_malloc_internal(nelem * elsize);
    //calloc returns zeroed memory by specification
    memset(p, '\0', nelem * elsize);
    return p;
}

/**
 * Allocates memory with ptmalloc2 and wraps object
 * header around requested memory.
 */
void *__wrap_malloc(size_t size) {

    object_header_t *object =
        (object_header_t*) (__real_malloc(size + sizeof (object_header_t)));

    if (!object) {
        perror("malloc");
        return NULL;
    }
    object->dc_or_region_id = 0;
    object->finalizer_index = -1;

#ifdef SCM_PRINTOVERHEAD
    inc_overhead(sizeof (object_header_t));
#endif

#ifdef SCM_PRINTMEM
    inc_allocated_mem(__real_malloc_usable_size(object));
    print_memory_consumption();
#endif
    return PAYLOAD_OFFSET(object);
}

/**
 * Reallocates memory with ptmalloc2 and wraps object
 * header around requested memory.
 */
void *__wrap_realloc(void *ptr, size_t size) {

    if (ptr == NULL) return __wrap_malloc_internal(size);
    //else: create new object
    object_header_t *new_object =
        (object_header_t*) __real_malloc(size + sizeof (object_header_t));

    if (!new_object) {
        perror("realloc");
        return NULL;
    }
    new_object->dc_or_region_id = 0;
    new_object->finalizer_index = -1;

#ifdef SCM_PRINTOVERHEAD
    inc_overhead(sizeof (object_header_t));
#endif

    //get the minimum of the old size and the new size
    size_t old_object_size =
        __real_malloc_usable_size(OBJECT_HEADER(ptr))
        - sizeof (object_header_t);
    size_t lesser_object_size;
    if (old_object_size >= size) {
        lesser_object_size = size;
    } else {
        lesser_object_size = old_object_size;
    }

    object_header_t *old_object = OBJECT_HEADER(ptr);
    //copy payload bytes 0..(lesser_size-1) from the old object to the new one
    memcpy(PAYLOAD_OFFSET(new_object),
           PAYLOAD_OFFSET(old_object),
           lesser_object_size);

    if (old_object->dc_or_region_id == 0) {
        //if the old object has no descriptors, we can free it
#ifdef SCM_PRINTMEM
        inc_freed_mem(__real_malloc_usable_size(old_object));
#endif

#ifdef SCM_PRINTOVERHEAD
        dec_overhead(sizeof (object_header_t));
#endif
        __real_free(old_object);
    } //else: the old object will be freed later due to expiration

#ifdef SCM_PRINTMEM
    inc_allocated_mem(__real_malloc_usable_size(new_object));
    print_memory_consumption();
#endif
    return PAYLOAD_OFFSET(new_object);
}

/**
 * Dellocates memory if object descriptor
 * counter is 0.
 */
void __wrap_free(void *ptr) {

    if (ptr == NULL) return;

    object_header_t *object = OBJECT_HEADER(ptr);

    if (object->dc_or_region_id == 0) {
#ifdef SCM_PRINTOVERHEAD
        dec_overhead(sizeof (object_header_t));
#endif
#ifdef SCM_PRINTMEM
        inc_freed_mem(__real_malloc_usable_size(object));
#endif
        __real_free(object);
    } else {
#ifdef SCM_DEBUG
        if(object->dc_or_region_id > 0) {
            printf("Cannot free objects which are still referenced.\n");
        } else if(object->dc_or_region_id < 0) {
            printf("Cannot free single objects from a region\n");
        }
#endif
    }
}

/**
 * Returns malloc usable size and
 * considers object header.
 */
size_t __wrap_malloc_usable_size(void *ptr) {

    object_header_t *object = OBJECT_HEADER(ptr);
    return __real_malloc_usable_size(object) - sizeof (object_header_t);
}

/**
 * scm_tick is used to advance the local time of the calling thread
 */
void scm_tick(void) {
    scm_tick_clock(0);
}

/**
 * scm_global_tick is used to advance the global time of the calling thread
 */
void scm_global_tick(void) {
    MICROBENCHMARK_START
    descriptor_root_t *descriptor_root = get_descriptor_root();

#ifdef SCM_DEBUG
    printf("scm_global_tick GT: %lu GP: %lu #T:%d ttc:%d\n",
           global_time, descriptor_root->global_phase,
           number_of_threads, ticked_threads_countdown);
#endif

    if (global_time == descriptor_root->global_phase) {

        //each thread must expire its own globally clocked buffer,
        //but can only do so on its first tick after the last global
        //time advance

        //my first tick in this global period
        descriptor_root->global_phase++;

        //current_index is equal to the so-called thread-global time
        increment_current_index(&descriptor_root->globally_clocked_obj_buffer);
        increment_current_index(&descriptor_root->globally_clocked_reg_buffer);

        //expire_buffer operates on current_index - 1, so it is called after
        //we incremented the current_index of the globally_clocked_buffer
        expire_buffer(&descriptor_root->globally_clocked_obj_buffer,
                      &descriptor_root->list_of_expired_obj_descriptors);
        expire_buffer(&descriptor_root->globally_clocked_reg_buffer,
                      &descriptor_root->list_of_expired_reg_descriptors);

        if (atomic_int_dec_and_test((int*) & ticked_threads_countdown)) {

            lock_global_time();

            ticked_threads_countdown = number_of_threads;
            //assert: descriptor_root->global_phase == global_time + 1
            global_time++;

            unlock_global_time();

        } //else global time does not advance

    } //else: we already ticked in this global_phase


	if(SCM_MAX_CLOCKS > 1) {

		unsigned int rr_index = descriptor_root->round_robin;

#ifdef SCM_CHECK_CONDITIONS
		if (rr_index == 0 || rr_index >= SCM_MAX_CLOCKS) {
			errno = EACCES;
			perror("The round robin index must never be 0 or >= SCM_MAX_CLOCKS");
			return;
		}
#endif
		unsigned int age_of_rr_buffer =
			descriptor_root->locally_clocked_obj_buffer[rr_index].age;

		// if the next round_robin buffer is a zombie -> cleanup incrementally
		if ( age_of_rr_buffer != descriptor_root->current_time &&
				descriptor_root->locally_clocked_obj_buffer[rr_index]
				.not_expired_length != 0) {

			increment_and_expire_clock(rr_index);

			rr_index = (rr_index + 1) % SCM_MAX_CLOCKS;
			if (rr_index == 0) {
				rr_index = 1;
			}
			descriptor_root->round_robin = rr_index;
		}
	}

#ifndef SCM_EAGER_COLLECTION
    //we also process expired descriptors at tick
    //to get a cyclic allocation/free scheme. this is optional
    scm_lazy_collect();
#else
    scm_eager_collect();
#endif

#ifdef SCM_PRINTMEM
    print_memory_consumption();
#endif
    MICROBENCHMARK_STOP
    MICROBENCHMARK_DURATION("scm_global_tick")
}

#ifndef SCM_EAGER_COLLECTION
/**
 * Collects descriptors incrementally
 */
void scm_lazy_collect(void) {

    descriptor_root_t *descriptor_root = get_descriptor_root();

    expire_obj_descriptor_if_exists(&descriptor_root->list_of_expired_obj_descriptors);
    expire_reg_descriptor_if_exists(&descriptor_root->list_of_expired_reg_descriptors);

}
#else
/**
 * Collects descriptors all at once
 */
void scm_eager_collect(void) {

    descriptor_root_t *descriptor_root = get_descriptor_root();

    while (expire_obj_descriptor_if_exists(
                &descriptor_root->list_of_expired_obj_descriptors));
    while (expire_reg_descriptor_if_exists(
                &descriptor_root->list_of_expired_reg_descriptors));

}
#endif

static inline int check_extension(int given_extension) {
    if (given_extension > SCM_MAX_EXPIRATION_EXTENSION) {
#ifdef SCM_DEBUG
        printf("violation of SCM_MAX_EXPIRATION_EXTENT\n");
#endif
        return SCM_MAX_EXPIRATION_EXTENSION;
    } else {
        return given_extension;
    }
}

/**
 * scm_global_refresh adds extension time units to the expiration time of
 * ptr and takes care that all other threads have enough time to also call
 * global_refresh(ptr, extension). If the object is part of a region, the
 * region is refreshed instead.
 */
void scm_global_refresh(void *ptr, unsigned int extension) {
    MICROBENCHMARK_START
#ifdef SCM_DEBUG
    printf("scm_global_refresh(%lx, %d)\n", (unsigned long) ptr, extension);
#endif

    descriptor_root_t *descriptor_root = get_descriptor_root();
    object_header_t *object = OBJECT_HEADER(ptr);

    extension = check_extension(extension);

    if (object->dc_or_region_id < 0) {
        int region_id = object->dc_or_region_id & ~HB_MASK;
        region_t* region = &(descriptor_root->regions[region_id]);
        atomic_int_inc((int*) & region->dc);

        insert_descriptor(region,
                          &descriptor_root->globally_clocked_reg_buffer, extension + 2);
    } else {
        atomic_int_inc((int*) & object->dc_or_region_id);

        insert_descriptor(object,
                          &descriptor_root->globally_clocked_obj_buffer, extension + 2);
    }


#ifndef SCM_EAGER_COLLECTION
    scm_lazy_collect();
#else
    //do nothing. expired descriptors are collected at tick
#endif

#ifdef SCM_PRINTMEM
    print_memory_consumption();
#endif
    MICROBENCHMARK_STOP
    MICROBENCHMARK_DURATION("scm_global_refresh")
}

/**
 * scm_refresh is the same as scm_global_refresh but does not take
 * care for other threads. In a multi-clock environment, scm_refresh refreshes
 * the object with the thread-local base clock.
 * If the object is part of a region, the region is refreshed instead.
 */
void scm_refresh(void *ptr, unsigned int extension) {
    scm_refresh_with_clock(ptr, extension, 0);
}

/**
 * get_descriptor_root is called by operations that need to access the thread
 * local meta data structures.
 */
descriptor_root_t *get_descriptor_root() {

    void *ptr = pthread_getspecific(descriptor_root_key);

    if (ptr == NULL) {
        ptr = scm_register_thread();
    }

    return (descriptor_root_t*) ptr;
}

/**
 * scm_register_thread is called on a thread when it operates the first time
 * in libscm. The thread data structures are created or reused from previously
 * terminated threads.
 */
static descriptor_root_t *scm_register_thread() {
    descriptor_root_t *descriptor_root;

    lock_descriptor_roots();

    if (init_threads == 0) {
        //initialize thread local data key and the global_time_lock spinlock
        //when the first thread is registered
        pthread_key_create(&descriptor_root_key, NULL);
        pthread_spin_init(&global_time_lock, PTHREAD_PROCESS_PRIVATE);
        init_threads = 1;
    }

    if (terminated_descriptor_roots != NULL) {
        descriptor_root = terminated_descriptor_roots;
        terminated_descriptor_roots = terminated_descriptor_roots->next;
    } else {
        descriptor_root = new_descriptor_root();
    }

    // The current time is used to distinguish from zombie descriptor
    // buffers which have another "age".
    // There will be an automatic buffer overflow if the last current_time
    // was equal to UINT_MAX
    descriptor_root->current_time++;

    int current_time = descriptor_root->current_time;
    descriptor_root->locally_clocked_obj_buffer[0].age = current_time;
    descriptor_root->locally_clocked_reg_buffer[0].age = current_time;
    unlock_descriptor_roots();

    pthread_setspecific(descriptor_root_key, descriptor_root);

    //assert: if descriptor_root belonged to a terminated thread,
    //block_thread was invoked on this thread
    scm_resume_thread_internal();

    return descriptor_root;
}

/**
 * scm_unregister_thread is called upon termination of a thread. The thread
 * leaves the system and passes its data structures in a pool to be reused
 * by other threads upon creation.
 */
void scm_unregister_thread() {

    scm_block_thread_internal();

    descriptor_root_t *descriptor_root = get_descriptor_root();

    lock_descriptor_roots();

    descriptor_root->next = terminated_descriptor_roots;
    terminated_descriptor_roots = descriptor_root;

    unlock_descriptor_roots();
}

/**
 * scm_block_thread is called when a thread blocks to notify the system about it
 */
void scm_block_thread() {
    descriptor_root_t *descriptor_root = get_descriptor_root();

    //assert: we do not have the descriptor_roots lock
    lock_global_time();
    number_of_threads--;

    if (global_time == descriptor_root->global_phase) {
        //we have not ticked in this global period

        if (atomic_int_dec_and_test((int*) & ticked_threads_countdown)) {
            //we are the last thread to tick and therefore need to tick globally
            if (number_of_threads == 0) {
                ticked_threads_countdown = 1;
            } else {
                ticked_threads_countdown = number_of_threads;
            }
            global_time++;
        }
    }
    unlock_global_time();
}

/**
 * scm_resume_thread is called when a thread returns from blocking state to
 * notify the system about it.
 */
void scm_resume_thread() {
    descriptor_root_t *descriptor_root = get_descriptor_root();

    //assert: we do not have the descriptor_roots lock
    lock_global_time();

    if (number_of_threads == 0) {
        /* if this is the first thread to resume/register,
         * then we have to tick to make
         * global progress, unless another thread registers
         * assert: ticked_threads_countdown == 1
         */
        descriptor_root->global_phase = global_time;
    } else {
        //else: we do not tick globally in the current global period
        //to avoid decrement of the ticked_threads_countdown
        descriptor_root->global_phase = global_time + 1;
    }
    number_of_threads++;

    unlock_global_time();
}

static descriptor_root_t* new_descriptor_root() {

    //allocate descriptor_root 0 initialized
    descriptor_root_t *descriptor_root =
        __real_calloc(1, sizeof (descriptor_root_t));

#ifdef SCM_PRINTOVERHEAD
    inc_overhead(__real_malloc_usable_size(descriptor_root));
#endif
#ifdef SCM_PRINTMEM
    inc_allocated_mem(__real_malloc_usable_size(descriptor_root));
#endif

    /* UPDATED 02.03.2011 to support region allocations */
    descriptor_root->globally_clocked_obj_buffer.not_expired_length =
        SCM_MAX_EXPIRATION_EXTENSION + 3;
    descriptor_root->globally_clocked_reg_buffer.not_expired_length =
        SCM_MAX_EXPIRATION_EXTENSION + 3;
    descriptor_root->locally_clocked_obj_buffer[0].not_expired_length =
        SCM_MAX_EXPIRATION_EXTENSION + 1;
    descriptor_root->locally_clocked_reg_buffer[0].not_expired_length =
        SCM_MAX_EXPIRATION_EXTENSION + 1;

    return descriptor_root;
}

static inline void lock_global_time() {
#ifdef SCM_PRINTLOCK
    if (pthread_spin_trylock(&global_time_lock)) {
        printf("thread %ld BLOCKS on global_time_lock\n", pthread_self());
        pthread_spin_lock(&global_time_lock);
    }
#else
    pthread_spin_lock(&global_time_lock);
#endif
}

static inline void unlock_global_time() {
    pthread_spin_unlock(&global_time_lock);
}

static inline void lock_descriptor_roots() {
#ifdef SCM_PRINTLOCK
    if (pthread_mutex_trylock(&terminated_descriptor_roots_mutex)) {
        printf("thread %ld BLOCKS on lock_descriptor_roots\n", pthread_self());
        pthread_mutex_lock(&terminated_descriptor_roots_mutex);
    }
#else
    pthread_mutex_lock(&terminated_descriptor_roots_mutex);
#endif
}

static inline void unlock_descriptor_roots() {
    pthread_mutex_unlock(&terminated_descriptor_roots_mutex);
}

/**********************************************************************/
/************** Region-based memory management allocator **************/
/******************* for Self-Collecting Mutators *********************/
/**********************************************************************/

/**
 * scm_create_region() returns a const integer representing a new region
 * if available and -1 otherwise. The new region is detected by scanning
 * the descriptor_root regions array for a region that
 * does not yet have any region_page. If such a region is found,
 * a region_page is created and initialized.
 */
const int scm_create_region() {
    descriptor_root_t* descriptor_root = get_descriptor_root();
    region_t* region = NULL;
    int i = 0;
    region = &(descriptor_root->regions[i]);
    while (region->firstPage != NULL) {
        // if the mutator calls scm_create_region() without allocating
        // memory into it, dc will still be 0. So if age != current_time
        // and dc == 0, we can reuse the region.
        if (region->age != descriptor_root->current_time
                && region->dc == 0) {
            return (const int) i;
        }
        i++;
        if (i >= SCM_MAX_REGIONS) {
            errno = ENOBUFS;
            perror("Region contingency exceeded.");
            return -1;
        }
        region = &descriptor_root->regions[i];
    }
    region->age = descriptor_root->current_time;
    region_page_t* page = init_region_page(region);
    region->firstPage = page;

// check post-conditions
#ifdef SCM_CHECK_CONDITIONS
    if (region == NULL
            || region->firstPage == NULL
            || region->lastPage == NULL
            || region->dc != 0
            || region->number_of_region_pages != 1 ) {
        perror("Region creation or initialization failed\n");
        exit(errno);
    }
#endif

    return (const int) i;
}

/**
 * scm_malloc_region allocates memory in a region.
 * It adds space for an object header to
 * the requested memory and initializes the
 * memory header.
 *
 * Every memory allocation request is aligned to
 * a word to effectively use cache lines.
 *
 * If the requested amount of memory is bigger than the
 * max region_page payload size, regmalloc returns NULL
 * and prints an error with error number EMSGSIZE.
 * If the region does not contain at least one
 * region_pages it was not correctly initialized and
 * scm_ammloc_region will return with error number
 * EFAULT.
 */
void* scm_malloc_in_region(size_t _size, const int region_index) {

#ifdef SCM_CHECK_CONDITIONS
    if (region_index < 0 || region_index >= SCM_MAX_REGIONS) {
        errno = ENOBUFS;
        perror("Region index is invalid.");
        exit(errno);
    }
#endif

    descriptor_root_t* descriptor_root = get_descriptor_root();
    region_t* region = &descriptor_root->regions[region_index];
    // check pre-conditions
#ifdef SCM_CHECK_CONDITIONS
    if (region == NULL) {
        perror("Cannot allocate into NULL region\n");
        return NULL;
    } else {
        if (region->firstPage == NULL || region->lastPage == NULL) {
            errno = EINVAL;
            perror("Region was not correctly initialized\n");
            exit(errno);
        }
    }
    region_t* invar_region = region;
#endif

    if (region->age != descriptor_root->current_time) {
        region->age = descriptor_root->current_time;
    }
    object_header_t* object = NULL;

    /* aligned size is calculated */
    unsigned long size_not_aligned = (unsigned long) (_size + sizeof(object_header_t));
    unsigned long size_aligned;
    unsigned long modulus = (unsigned long) (size_not_aligned % (unsigned long) sizeof(long));
    if (modulus != 0) {
        size_aligned = ((size_not_aligned / sizeof(long)) + 1) * sizeof(long);
    } else {
        size_aligned = size_not_aligned;
    }

    // check if the requested size fits into a region page
    if (size_aligned > REGION_PAGE_PAYLOAD_SIZE) {
        errno = EMSGSIZE;
        printf("Size not aligned: %lu, size aligned: %lu\n", size_not_aligned, size_aligned);
        perror("The region allocator does not support memory of this size\n");
        exit(errno);
    }

    region_page_t* page = region->lastPage;

    // region page is filled up ... init another page
    if ((size_aligned + page->used_memory) > REGION_PAGE_PAYLOAD_SIZE) {

#ifdef SCM_DEBUG
        printf("Page is full: %lu bytes allocated\n Creating new page...", (unsigned long) page->used_memory);
        printf("[new_region_page (%u)]\n", REGION_PAGE_SIZE);
#endif
        // allocate new page
        page = init_region_page(region);
    }
    // set the pointer to the allocated memory
    object = (object_header_t*) &page->memory[page->used_memory];
    // update amount of used memory
    page->used_memory += size_aligned;

// check post-conditions
#ifdef SCM_CHECK_CONDITIONS
    if (region != invar_region) {
        errno = EINVAL;
        perror("The region or the first region-page changed during initialization\n");
        exit(errno);
    }
    if (object == NULL) {
        errno = EINVAL;
        perror("Error during allocation. Object is NULL\n");
        exit(errno);
    }
    unsigned long not_word_aligned = (unsigned long) (page->used_memory % (unsigned long) sizeof(long));
    if (not_word_aligned) {
        errno = EINVAL;
        perror("Requested memory was not word aligned\n");
        exit(errno);
    }
#endif

    object->dc_or_region_id = region_index | HB_MASK;
    object->finalizer_index = -1;

    return PAYLOAD_OFFSET(object);
}

/**
 * Creates and initializes a new region page if no other
 * region page exists or if all other region pages are full.
 * The region_page is allocated page-aligned.
 */
static region_page_t* init_region_page(region_t* region) {

    descriptor_root_t* descriptor_root = get_descriptor_root();
// check pre-conditions
#ifdef SCM_CHECK_CONDITIONS
    if (region == NULL) {
        errno = ECANCELED;
        perror("Cannot initialize region_page for NULL region\n");
        exit(errno);
    } else if (region->age != descriptor_root->current_time) {
        errno = ECANCELED;
        perror("Initializing region page into zombie region is not allowed\n");
        exit(errno);
    }
    region_t* invar_region = region;
    region_page_t* invar_first_region_page = region->firstPage;
#endif

    region_page_t* prevLastPage = region->lastPage;

    region_page_t* new_page = descriptor_root->region_page_pool;
    if (new_page != NULL) {
        descriptor_root->region_page_pool = new_page->nextPage;
        descriptor_root->number_of_pooled_region_pages--;
#ifdef SCM_PRINTMEM
        dec_pooled_mem(sizeof (region_page_t));
#endif
    } else {

        int bytes = posix_memalign((void**) &new_page, sysconf(_SC_PAGESIZE), REGION_PAGE_SIZE);

#ifdef SCM_PRINTMEM
        printf("(%lx) allocated bytes: %lu\n", (unsigned long) region,bytes);
        inc_allocated_mem(bytes);
#endif

        if (new_page == NULL) {
            perror("memory could not be allocated or aligned\n");
            return NULL;
        }
    }
    memset(new_page, 0, REGION_PAGE_SIZE);

    if (prevLastPage != NULL) {
        prevLastPage->nextPage = new_page;
    }

    region->lastPage = new_page;
    region->number_of_region_pages++;

// check post-conditions
#ifdef SCM_CHECK_CONDITIONS
    if (region == NULL) {
        errno = ECANCELED;
        perror("The region became NULL during initialization of a region page\n");
        exit(errno);
    } else if (region != invar_region || region->firstPage != invar_first_region_page) {
        errno = ECANCELED;
        perror("The region or the first region-page changed during initialization\n");
        exit(errno);
    } else if (new_page == NULL || new_page->nextPage != NULL || new_page->used_memory != 0) {
        errno = ECANCELED;
        perror("The new region page was not correctly initialized\n");
        exit(errno);
    }
#endif

#ifdef SCM_METER
    add_allocated_mem(&new_page, REGION_PAGE_SIZE, 0);
#endif
    return new_page;
}

/**
 * scm_create_clock() returns a const integer representing
 * a new clock in the short-term memory model.
 * A clock identifies a descriptor buffer in the array of locally
 * clocked descriptor buffers of the descriptor root.
 * If all available clocks/descriptor buffers are in use, the return value is
 * set to -1, indicating an error for the caller function.
 */
const int scm_register_clock() {
    descriptor_root_t* descriptor_root = get_descriptor_root();
    int i = 1;

    while (descriptor_root->locally_clocked_obj_buffer[i].age ==
            descriptor_root->current_time) {
        if (i > SCM_MAX_CLOCKS) {
            errno = ENOBUFS;
            perror("Clock contingency exceeded.\n");
            return -1;
        }
        i++;
    }
    descriptor_root->locally_clocked_obj_buffer[i].not_expired_length =
        SCM_MAX_EXPIRATION_EXTENSION + 1;
    descriptor_root->locally_clocked_reg_buffer[i].not_expired_length =
        SCM_MAX_EXPIRATION_EXTENSION + 1;

    descriptor_root->locally_clocked_obj_buffer[i].age =
        descriptor_root->current_time;
    descriptor_root->locally_clocked_reg_buffer[i].age =
        descriptor_root->current_time;

    return (const int) i;
}

/**
 * scm_unregister_clock() sets the age back to a value that is not equal
 * to the descriptor_root current_time. As a consequence the clock buffer
 * will be cleaned up incrementally during scm_tick() calls.
 */
void scm_unregister_clock(const int clock) {

#ifdef SCM_CHECK_CONDITIONS
    if (clock <= 1 || clock >= SCM_MAX_CLOCKS) {
        perror("Clock index is invalid.");
        return;
    }
#endif

    descriptor_root_t* descriptor_root = get_descriptor_root();

    descriptor_root->locally_clocked_obj_buffer[clock].age =
        (descriptor_root->current_time - 1);
    descriptor_root->locally_clocked_reg_buffer[clock].age =
        (descriptor_root->current_time - 1);
}

/**
 * scm_unregister_region() sets the age back to a value that is not equal
 * to the descriptor_root current_time. As a consequence the region may
 * be reused again if the dc is 0.
 */
void scm_unregister_region(const int region) {

#ifdef SCM_CHECK_CONDITIONS
    if (region <= 0 || region >= SCM_MAX_REGIONS) {
        perror("Region index is invalid.");
        return;
    }
#endif

    descriptor_root_t* descriptor_root = get_descriptor_root();

    descriptor_root->regions[region].age =
        (descriptor_root->current_time - 1);
    descriptor_root->regions[region].age =
        (descriptor_root->current_time - 1);
}

/**
 * increment_and_expire increments the current index of
 * the locally clocked descriptor buffers
 * and expires the descriptors from the last index
 */
void increment_and_expire_clock(const unsigned long clock) {
    descriptor_root_t *descriptor_root = get_descriptor_root();

#ifdef SCM_CHECK_CONDITIONS
    if (clock < 0 || clock >= SCM_MAX_CLOCKS) {
        errno = EINVAL;
        perror("The given clock is out of range.");
        exit(errno);
    }
#endif
    //make local time progress
    //current_index is equal to the so-called thread-local time
    increment_current_index(
        &descriptor_root->locally_clocked_obj_buffer[clock]);
    increment_current_index(
        &descriptor_root->locally_clocked_reg_buffer[clock]);

    //expire_buffer operates on current_index - 1, so it is called after
    //we incremented the current_index of the locally_clocked_buffer
    expire_buffer(&descriptor_root->locally_clocked_obj_buffer[clock],
                  &descriptor_root->list_of_expired_obj_descriptors);
    expire_buffer(&descriptor_root->locally_clocked_reg_buffer[clock],
                  &descriptor_root->list_of_expired_reg_descriptors);
}

/**
 * scm_tick_clock is used to advance the time of the given thread-local clock
 */
void scm_tick_clock(const unsigned long clock) {
    MICROBENCHMARK_START

#ifdef SCM_DEBUG
    printf("Ticking clock: %u\n", clock);
#endif
    increment_and_expire_clock(clock);

    descriptor_root_t *descriptor_root = get_descriptor_root();
    
	if(SCM_MAX_CLOCKS > 1) {
		unsigned int rr_index = descriptor_root->round_robin;
		if (rr_index == clock) {
			rr_index = (rr_index + 1) % SCM_MAX_CLOCKS;
			if (rr_index == 0) {
				rr_index = 1;
			}
		}

#ifdef SCM_CHECK_CONDITIONS
		if (rr_index == 0 || rr_index >= SCM_MAX_CLOCKS) {
			errno = EACCES;
			perror("The round robin index must never be 0 or >= SCM_MAX_CLOCKS");
			return;
		}
#endif
		unsigned int age_of_rr_buffer =
			descriptor_root->locally_clocked_obj_buffer[rr_index].age;

		// if the next round_robin buffer is a zombie -> cleanup incrementally
		if ( age_of_rr_buffer != descriptor_root->current_time &&
            descriptor_root->locally_clocked_obj_buffer[rr_index]
				.not_expired_length != 0) {

			increment_and_expire_clock(rr_index);

			rr_index = (rr_index + 1) % SCM_MAX_CLOCKS;
			if (rr_index == 0) {
				rr_index = 1;
			}
			descriptor_root->round_robin = rr_index;
		}
	}
#ifndef SCM_EAGER_COLLECTION
    //we also process expired descriptors at tick
    //to get a cyclic allocation/free scheme. this is optional
    scm_lazy_collect();
#else
    scm_eager_collect();
#endif

#ifdef SCM_PRINTMEM
    print_memory_consumption();
#endif

    MICROBENCHMARK_STOP
    MICROBENCHMARK_DURATION("scm_tick")
}

/**
 * scm_refresh_region() adds extension time units to the expiration time of
 * a region without taking care of other threads.
 * In a multi-clock environment, scm_refresh refreshes
 * the region with the thread-local base clock.
 */
void scm_refresh_region(const int region_id, unsigned int extension) {
	scm_refresh_region_with_clock(region_id, extension, 0);
}

/**
 * scm_refresh_with_clock refreshes a given object with a given clock,
 * which can be different to the thread-local base clock.
 * If an object is refreshed with multiple clocks it lives
 * until all clocks ticked n times, where n is the respective extension.
 * If the object is part of a region, the region is refreshed instead.
 */
void scm_refresh_with_clock(void *ptr, unsigned int extension, const unsigned long clock) {
    MICROBENCHMARK_START
    object_header_t *object = OBJECT_HEADER(ptr);

    /************ adapted for region-based allocator ************/
    // is the object allocated into a region?
    if (object->dc_or_region_id < 0) {
        int region_id = object->dc_or_region_id & ~HB_MASK;
        scm_refresh_region_with_clock(region_id, extension, clock);
    } else {

    	descriptor_root_t *descriptor_root = get_descriptor_root();
    	extension = check_extension(extension);

#ifdef SCM_DEBUG
    printf("scm_refresh(%lx, %d)\n", (unsigned long) ptr, extension);
#endif

// check pre-conditions
#ifdef SCM_CHECK_CONDITIONS
    if (ptr == NULL) {
        perror("Cannot refresh NULL pointer\n");
        return;
    }
#endif

#ifdef SCM_CHECK_CONDITIONS
        if (descriptor_root->current_time !=
                descriptor_root->locally_clocked_reg_buffer[clock].age ||
                descriptor_root->locally_clocked_reg_buffer[clock]
                .not_expired_length == 0) {
            perror("Cannot refresh zombie clock");
            return;
        }
#endif
        if (object->dc_or_region_id == INT_MAX) {
            perror("Descriptor counter reached max value");
            return;
        }
        atomic_int_inc((int*) & object->dc_or_region_id);
        insert_descriptor(object,
                          &descriptor_root->locally_clocked_obj_buffer[clock], extension);
    }

#ifndef SCM_EAGER_COLLECTION
    scm_lazy_collect();
#else
    //do nothing. expired descriptors are collected at tick
#endif
    /************ END: adapted for region-based allocator ************/

#ifdef SCM_PRINTMEM
    print_memory_consumption();
#endif
    MICROBENCHMARK_STOP
    MICROBENCHMARK_DURATION("scm_refresh")
}


/**
 * scm_refresh_region_with_clock() refreshes a given region with a given
 * clock, which can be different from the thread-local base clock.
 * If a region is refreshed with multiple clocks it lives
 * until all clocks ticked n times, where n is the respective extension.
 */
void scm_refresh_region_with_clock(const int region_id, unsigned int extension, const unsigned long clock) {
    MICROBENCHMARK_START
#ifdef SCM_DEBUG
    printf("scm_refresh_region_with_clock(%d, %u, %u)\n", region_id, extension, clock);
#endif

// check pre-conditions
#ifdef SCM_CHECK_CONDITIONS
    if (region_id < 0 || region_id > SCM_MAX_REGIONS) {
        perror("Region id is out of range\n");
        return;
    }
#endif

    descriptor_root_t *descriptor_root = get_descriptor_root();
    extension = check_extension(extension);

#ifdef SCM_DEBUG
	printf("region id: %d\n", region_id);
#endif
	region_t* region = &(descriptor_root->regions[region_id]);
#ifdef SCM_CHECK_CONDITIONS
	if (descriptor_root->current_time !=
			descriptor_root->locally_clocked_reg_buffer[clock].age ||
			descriptor_root->locally_clocked_reg_buffer[clock]
			.not_expired_length == 0) {
		perror("Cannot refresh zombie or uninitialized clock");
		return;
	}
#endif
	atomic_int_inc((int*) & region->dc);
	insert_descriptor(region,
					  &descriptor_root->locally_clocked_reg_buffer[clock], extension);


#ifndef SCM_EAGER_COLLECTION
    scm_lazy_collect();
#else
    //do nothing. expired descriptors are collected at tick
#endif
    /************ END: adapted for region-based allocator ************/

#ifdef SCM_PRINTMEM
    print_memory_consumption();
#endif
    MICROBENCHMARK_STOP
    MICROBENCHMARK_DURATION("scm_refresh")
}

